#include "project.h"

/*Colormap is 565 RGB (B is least significant) */


#define TX 440
#define TY 326
#define MT 20

static int tx = TX, ty = TY;
static float tdiv = 1;

static int maxtiles = MT;

static IOSMAP current_map;


BEGIN_OBJECT(Tile)
IOSMAP m;
Pixmap p;
int depth;
END_OBJECT(Tile) CREATOR(Tile) = (Creator(Tile)) Obj_DefaultCreator;
static void tile_destructor(Tile);
DESTRUCTOR(Tile) = tile_destructor;

static Tile_list tiles = NULL;
static int numtiles = 0;

static void tile_destructor(Tile t)
{
  XFreePixmap(dpy, t->p);
  free(t);
}


static int org_lx()
{
  int w = xv_get(canvas, CANVAS_WIDTH);
  return (w - tx) >> 1;

}

static int org_ty()
{
  int h = xv_get(canvas, CANVAS_HEIGHT);
  return ((h - ty) >> 1);
}


static Tile load_tile_by_iosmap(IOSMAP m)
{
  Tile ret;
  char *fn;
  GdkPixbuf *pb, *pbs;

  fn = iosmap_to_filename(m);
  pb = gdk_pixbuf_new_from_file(fn);
  free(fn);

  if (!pb)
    return NULL;

  ret = CREATE_OBJ(Tile);
  ret->m = m;
  ret->depth = colordepth;

  ret->p = XCreatePixmap(dpy, win, tx, ty, colordepth);

  pbs = gdk_pixbuf_scale_simple(pb, tx, ty, GDK_INTERP_BILINEAR);
  gdk_pixbuf_finalize(pb);

  gdk_pixbuf_xlib_render_to_drawable(pbs, ret->p, gc,
                                     0, 0, 0, 0, tx, ty,
                                     XLIB_RGB_DITHER_NORMAL, 0, 0);

  gdk_pixbuf_finalize(pbs);
  return ret;
}

static Tile get_tile_by_iosmap(IOSMAP m)
{
  Tile t;

  if (!tiles)
    tiles = CREATE_LIST(Tile);
  t = tiles->head;

  while (t) {
    if (ios_same_map(t->m, m)) {
      Obj_Remove(tiles, t);
      Obj_Insert(tiles, t, ObjInsertTail);
      return t;
    }
    t = t->next;
  }

  if (numtiles >= maxtiles) {
    t = tiles->head;
    Obj_Remove(tiles, t);
    Obj_Delete(t);
  }

  t = load_tile_by_iosmap(m);
  if (t)
    Obj_Insert(tiles, t, ObjInsertTail);

  return t;
}

static void flush_tile_cache()
{
  if (tiles)
    Obj_Delete(tiles);
  tiles = NULL;
}


static int draw_cursor()
{
  int w, h, cx, cy;
  IOSGRID g = wgs84_to_iosgrid(pos);
  IOSXY xy = iosgrid_to_iosxy(g);
  IOSMAP m = iosxy_to_iosmap(xy);


  cx = org_lx();
  cy = org_ty();
  cy += ty;

  cx += (int) (m.x * (float) tx);
  cy -= (int) (m.y * (float) ty);

  cursor_put(cx, cy, heading);

  return (!ios_same_map(m, current_map));
}

void i_canvas_repaint(void)
{
  int i;
  int w, h, x, y;
  int ne, nn, n, e;
  IOSGRID g = wgs84_to_iosgrid(pos);
  IOSXY nxy, xy = iosgrid_to_iosxy(g);
  IOSMAP m;

  Tile t;

  w = xv_get(canvas, CANVAS_WIDTH);
  h = xv_get(canvas, CANVAS_HEIGHT);

  XSetForeground(dpy, gc, 0x0);

  x = org_lx();
  y = org_ty();

  ne = x / tx;
  ne += 2;

  nn = y / ty;
  nn += 2;

  for (n = -nn; n <= nn; ++n) {
    for (e = -ne; e <= ne; ++e) {
      int lx = x + (e * tx);
      int ly = y - (n * ty);

      if ((lx > -tx) && (lx < w) && (ly > -ty) && (ly < h)) {
        nxy = iosxy_neighbour(xy, e, n);
        m = iosxy_to_iosmap(nxy);

        if ((!e) && (!n))
          current_map = m;
/*
printf("Fetching... %d %d ...",x+(e*tx),y-(n*ty));
print_bosmap(m);
*/
        t = get_tile_by_iosmap(m);
        if (t)
          XCopyArea(dpy, t->p, win, gc, 0, 0, tx, ty, lx, ly);
        else
          XFillRectangle(dpy, win, gc, lx, ly, tx, ty);
      }
    }
  }


  cursor_redraw();
  draw_cursor();
}


void i_set_tdiv(float q)
{

  tdiv = q;
  tx = ((float) TX) / q;
  ty = ((float) TY) / q;
  maxtiles = ((int) ((tdiv + .5) * (tdiv + .5))) * MT;
  maxtiles++;

  flush_tile_cache();
  i_canvas_repaint();
}

float i_get_size(void)
{
  float q = (float) tx;         /*Size of tile in pixels */
  q = q / 7800.0;               /*Size of tile in meters */

  return q;
}

void i_set_size(float q)
{
  int i;
  q = q * 7800;
  q = ((float) TX) / q;

  i_set_tdiv(q);

}


void i_flush_tiles(void)
{
  flush_tile_cache();
}

void i_new_pos()
{
  IOSGRID g = wgs84_to_iosgrid(pos);
  char buf[100];

  sprint_iosgrid(buf, g);

  xv_set(gridref, PANEL_VALUE, buf, 0);

  if (draw_cursor())
    i_canvas_repaint();

  printf("Scale is %f meters/pixel\n", 1.0 / i_get_size());
}





void i_clicked(int x, int y, int move)
{
  int w, h, cx, cy;
  IOSGRID g2, g = wgs84_to_iosgrid(pos);
  IOSXY xy = iosgrid_to_iosxy(g);
  IOSMAP m = iosxy_to_iosmap(xy);
  IOSXY xy_bl, xy_tr;
  float fx, fy;
  char buf[100];

  m.x = m.y = 0.0;
  xy_bl = iosmap_to_iosxy(m);

  m.x = m.y = 1.0;
  xy_tr = iosmap_to_iosxy(m);

  cx = org_lx();
  cy = org_ty();
  cy += ty;

  x -= cx;
  y = cy - y;

  fx = (float) (xy_tr.x - xy_bl.x);
  fy = (float) (xy_tr.y - xy_bl.y);

  fx = fx / (float) tx;
  fy = fy / (float) ty;


  xy_bl.x += (int) ((fx * (float) x) + .5);
  xy_bl.y += (int) ((fx * (float) y) + .5);

  if (move) {
    g2 = iosxy_to_iosgrid(xy_bl);
    pos = iosgrid_to_wgs84(g2);
    new_pos();



  }

  xy.x -= xy_bl.x;
  xy.y -= xy_bl.y;

  fx = xy.x;
  fy = xy.y;
  fx = fx * fx + fy * fy;
  fx = sqrt(fx);
  fy = fx / 1000.f;

  sprintf(buf, "%.3f km", fy);
  xv_set(dist, PANEL_VALUE, buf, 0);

  if (speed_mps > 0.00001)
    fx = fx / speed_mps;
  else
    fx = 0;


  {
    int h, m, s;

    s = (int) (fx + .5f);
    m = s / 60;
    s -= m * 60;
    h = m / 60;
    m -= h * 60;

    sprintf(buf, "%d:%02d:%02d", h, m, s);
    xv_set(eta, PANEL_VALUE, buf, 0);
  }

  m = iosxy_to_iosmap(xy_bl);

  g = iosxy_to_iosgrid(xy_bl);

  sprint_iosgrid(buf, g);

  xv_set(destref, PANEL_VALUE, buf, 0);

}
